"""Test writing sensor_msgs with pybag and reading with rosbags."""
from pathlib import Path
from tempfile import TemporaryDirectory

from mcap.reader import make_reader
from mcap_ros2.decoder import DecoderFactory

import pybag.ros2.humble.builtin_interfaces as builtin_interfaces
import pybag.ros2.humble.geometry_msgs as geometry_msgs
import pybag.ros2.humble.sensor_msgs as sensor_msgs
import pybag.ros2.humble.std_msgs as std_msgs
from pybag.mcap_writer import McapFileWriter


def _make_header(frame_id: str = 'frame', sec: int = 1, nanosec: int = 2) -> std_msgs.Header:
    return std_msgs.Header(
        stamp=builtin_interfaces.Time(sec=sec, nanosec=nanosec),
        frame_id=frame_id,
    )


def _write_pybag(temp_dir: str, msg, topic: str = '/pybag', *, timestamp: int = 0) -> Path:
    mcap_path = Path(temp_dir) / 'data.mcap'
    with McapFileWriter.open(mcap_path) as writer:
        writer.write_message(topic, timestamp, msg)
    return mcap_path


def _roundtrip_write(msg) -> list:
    with TemporaryDirectory() as temp_dir:
        path = _write_pybag(temp_dir, msg)
        with open(path, 'rb') as f:
            reader = make_reader(f, decoder_factories=[DecoderFactory()])
            return [ros_msg for _, _, _, ros_msg in reader.iter_decoded_messages()]


# Tests ---------------------------------------------------------------------

def test_battery_state_pybag() -> None:
    msg = sensor_msgs.BatteryState(
        header=_make_header(),
        voltage=12.5,
        temperature=25.0,
        current=1.5,
        charge=100.0,
        capacity=200.0,
        design_capacity=200.0,
        percentage=50.0,
        power_supply_status=sensor_msgs.BatteryState.POWER_SUPPLY_STATUS_CHARGING,
        power_supply_health=sensor_msgs.BatteryState.POWER_SUPPLY_HEALTH_GOOD,
        power_supply_technology=sensor_msgs.BatteryState.POWER_SUPPLY_TECHNOLOGY_LION,
        present=True,
        cell_voltage=[3.0, 3.0, 3.0],
        cell_temperature=[25.0, 25.0, 25.0],
        location='main_battery',
        serial_number='BAT001'
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'BatteryState'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].voltage == 12.5
    assert decoded_msgs[0].temperature == 25.0
    assert decoded_msgs[0].current == 1.5
    assert decoded_msgs[0].charge == 100.0
    assert decoded_msgs[0].capacity == 200.0
    assert decoded_msgs[0].design_capacity == 200.0
    assert decoded_msgs[0].percentage == 50.0
    assert decoded_msgs[0].power_supply_status == 1      # POWER_SUPPLY_STATUS_CHARGING
    assert decoded_msgs[0].power_supply_health == 1      # POWER_SUPPLY_HEALTH_GOOD
    assert decoded_msgs[0].power_supply_technology == 2  # POWER_SUPPLY_TECHNOLOGY_LION
    assert decoded_msgs[0].present is True
    assert decoded_msgs[0].cell_voltage == [3.0, 3.0, 3.0]
    assert decoded_msgs[0].cell_temperature == [25.0, 25.0, 25.0]
    assert decoded_msgs[0].location == 'main_battery'
    assert decoded_msgs[0].serial_number == 'BAT001'


def test_region_of_interest_pybag() -> None:
    msg = sensor_msgs.RegionOfInterest(
        x_offset=10,
        y_offset=20,
        height=100,
        width=200,
        do_rectify=True
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'RegionOfInterest'
    assert decoded_msgs[0].x_offset == 10
    assert decoded_msgs[0].y_offset == 20
    assert decoded_msgs[0].height == 100
    assert decoded_msgs[0].width == 200
    assert decoded_msgs[0].do_rectify is True


def test_camera_info_pybag() -> None:
    msg = sensor_msgs.CameraInfo(
        header=_make_header(),
        height=480,
        width=640,
        distortion_model='plumb_bob',
        d=[0.1, 0.2, 0.0, 0.0, 0.0],
        k=[1000.0, 0.0, 320.0, 0.0, 1000.0, 240.0, 0.0, 0.0, 1.0],
        r=[1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
        p=[1000.0, 0.0, 320.0, 0.0, 0.0, 1000.0, 240.0, 0.0, 0.0, 0.0, 1.0, 0.0],
        binning_x=1,
        binning_y=1,
        roi=sensor_msgs.RegionOfInterest(
            x_offset=0,
            y_offset=0,
            height=480,
            width=640,
            do_rectify=False
        )
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'CameraInfo'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].height == 480
    assert decoded_msgs[0].width == 640
    assert decoded_msgs[0].distortion_model == 'plumb_bob'
    assert decoded_msgs[0].d == [0.1, 0.2, 0.0, 0.0, 0.0]
    assert decoded_msgs[0].k == [1000.0, 0.0, 320.0, 0.0, 1000.0, 240.0, 0.0, 0.0, 1.0]
    assert decoded_msgs[0].r == [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0]
    assert decoded_msgs[0].p == [1000.0, 0.0, 320.0, 0.0, 0.0, 1000.0, 240.0, 0.0, 0.0, 0.0, 1.0, 0.0]
    assert decoded_msgs[0].binning_x == 1
    assert decoded_msgs[0].binning_y == 1
    assert decoded_msgs[0].roi.x_offset == 0
    assert decoded_msgs[0].roi.y_offset == 0
    assert decoded_msgs[0].roi.height == 480
    assert decoded_msgs[0].roi.width == 640
    assert decoded_msgs[0].roi.do_rectify is False


def test_channel_float32_pybag() -> None:
    msg = sensor_msgs.ChannelFloat32(
        name='intensity',
        values=[1.0, 2.0]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'ChannelFloat32'
    assert decoded_msgs[0].name == 'intensity'
    assert decoded_msgs[0].values == [1.0, 2.0]


def test_compressed_image_pybag() -> None:
    msg = sensor_msgs.CompressedImage(
        header=_make_header(),
        format='jpeg',
        data=[255, 216, 255, 224, 0, 16, 74, 70, 73, 70, 0, 1]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'CompressedImage'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].format == 'jpeg'
    assert decoded_msgs[0].data == bytes([255, 216, 255, 224, 0, 16, 74, 70, 73, 70, 0, 1])


def test_fluid_pressure_pybag() -> None:
    msg = sensor_msgs.FluidPressure(
        header=_make_header(),
        fluid_pressure=101325.0,
        variance=1.0
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'FluidPressure'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].fluid_pressure == 101325.0
    assert decoded_msgs[0].variance == 1.0


def test_illuminance_pybag() -> None:
    msg = sensor_msgs.Illuminance(
        header=_make_header(),
        illuminance=500.0,
        variance=0.1
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'Illuminance'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].illuminance == 500.0
    assert decoded_msgs[0].variance == 0.1


def test_image_pybag() -> None:
    msg = sensor_msgs.Image(
        header=_make_header(),
        height=480,
        width=640,
        encoding='rgb8',
        is_bigendian=0,
        step=1920,
        data=[255, 0, 0]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'Image'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].height == 480
    assert decoded_msgs[0].width == 640
    assert decoded_msgs[0].encoding == 'rgb8'
    assert decoded_msgs[0].is_bigendian == 0
    assert decoded_msgs[0].step == 1920
    assert decoded_msgs[0].data == bytes([255, 0, 0])


def test_imu_pybag() -> None:
    orientation = geometry_msgs.Quaternion(x=0.0, y=0.0, z=0.0, w=1.0)
    angular_velocity = geometry_msgs.Vector3(x=0.1, y=0.2, z=0.3)
    linear_acceleration = geometry_msgs.Vector3(x=9.8, y=0.0, z=0.0)
    msg = sensor_msgs.Imu(
        header=_make_header(),
        orientation=orientation,
        orientation_covariance=[1.0] * 9,
        angular_velocity=angular_velocity,
        angular_velocity_covariance=[0.1] * 9,
        linear_acceleration=linear_acceleration,
        linear_acceleration_covariance=[0.5] * 9
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'Imu'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].orientation.x == 0.0
    assert decoded_msgs[0].orientation.y == 0.0
    assert decoded_msgs[0].orientation.z == 0.0
    assert decoded_msgs[0].orientation.w == 1.0
    assert decoded_msgs[0].orientation_covariance == [1.0] * 9
    assert decoded_msgs[0].angular_velocity.x == 0.1
    assert decoded_msgs[0].angular_velocity.y == 0.2
    assert decoded_msgs[0].angular_velocity.z == 0.3
    assert decoded_msgs[0].angular_velocity_covariance == [0.1] * 9
    assert decoded_msgs[0].linear_acceleration.x == 9.8
    assert decoded_msgs[0].linear_acceleration.y == 0.0
    assert decoded_msgs[0].linear_acceleration.z == 0.0
    assert decoded_msgs[0].linear_acceleration_covariance == [0.5] * 9


def test_joint_state_pybag() -> None:
    msg = sensor_msgs.JointState(
        header=_make_header(),
        name=['joint1', 'joint2', 'joint3'],
        position=[0.5, 1.0, -0.3],
        velocity=[0.1, 0.2, 0.0],
        effort=[10.0, 5.0, 2.0]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'JointState'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].name == ['joint1', 'joint2', 'joint3']
    assert decoded_msgs[0].position == [0.5, 1.0, -0.3]
    assert decoded_msgs[0].velocity == [0.1, 0.2, 0.0]
    assert decoded_msgs[0].effort == [10.0, 5.0, 2.0]


def test_joy_pybag() -> None:
    msg = sensor_msgs.Joy(
        header=_make_header(),
        axes=[0.5, -0.5, 0.0, 1.0, 0.0, 0.0],
        buttons=[0, 1, 0, 0, 1, 0, 0, 0]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'Joy'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].axes == [0.5, -0.5, 0.0, 1.0, 0.0, 0.0]
    assert decoded_msgs[0].buttons == [0, 1, 0, 0, 1, 0, 0, 0]


def test_joy_feedback_pybag() -> None:
    msg = sensor_msgs.JoyFeedback(
        type=sensor_msgs.JoyFeedback.TYPE_LED,
        id=1,
        intensity=0.5
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'JoyFeedback'
    assert decoded_msgs[0].type == 0  # TYPE_LED
    assert decoded_msgs[0].id == 1
    assert decoded_msgs[0].intensity == 0.5


def test_joy_feedback_array_pybag() -> None:
    feedback1 = sensor_msgs.JoyFeedback(
        type=sensor_msgs.JoyFeedback.TYPE_LED,
        id=1,
        intensity=0.5
    )
    feedback2 = sensor_msgs.JoyFeedback(
        type=sensor_msgs.JoyFeedback.TYPE_RUMBLE,
        id=2,
        intensity=0.5
    )
    msg = sensor_msgs.JoyFeedbackArray(
        array=[feedback1, feedback2]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'JoyFeedbackArray'
    assert len(decoded_msgs[0].array) == 2
    assert decoded_msgs[0].array[0].type == 0  # TYPE_LED
    assert decoded_msgs[0].array[0].id == 1
    assert decoded_msgs[0].array[0].intensity == 0.5
    assert decoded_msgs[0].array[1].type == 1  # TYPE_RUMBLE
    assert decoded_msgs[0].array[1].id == 2
    assert decoded_msgs[0].array[1].intensity == 0.5


def test_laser_echo_pybag() -> None:
    msg = sensor_msgs.LaserEcho(
        echoes=[1.5, 2.0, 3.5, 4.0]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'LaserEcho'
    assert decoded_msgs[0].echoes == [1.5, 2.0, 3.5, 4.0]


def test_laser_scan_pybag() -> None:
    msg = sensor_msgs.LaserScan(
        header=_make_header(),
        angle_min=-1.5,
        angle_max=1.5,
        angle_increment=0.5,
        time_increment=0.25,
        scan_time=0.5,
        range_min=0.5,
        range_max=10.0,
        ranges=[1.0, 1.0, 1.0, 1.0, 1.0],
        intensities=[100.0, 110.0, 120.0, 130.0, 140.0]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'LaserScan'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].angle_min == -1.5
    assert decoded_msgs[0].angle_max == 1.5
    assert decoded_msgs[0].angle_increment == 0.5
    assert decoded_msgs[0].time_increment == 0.25
    assert decoded_msgs[0].scan_time == 0.5
    assert decoded_msgs[0].range_min == 0.5
    assert decoded_msgs[0].range_max == 10.0
    assert decoded_msgs[0].ranges == [1.0, 1.0, 1.0, 1.0, 1.0]
    assert decoded_msgs[0].intensities == [100.0, 110.0, 120.0, 130.0, 140.0]


def test_magnetic_field_pybag() -> None:
    magnetic_field = geometry_msgs.Vector3(x=0.1, y=0.2, z=0.3)
    msg = sensor_msgs.MagneticField(
        header=_make_header(),
        magnetic_field=magnetic_field,
        magnetic_field_covariance=[0.01] * 9
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'MagneticField'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].magnetic_field.x == 0.1
    assert decoded_msgs[0].magnetic_field.y == 0.2
    assert decoded_msgs[0].magnetic_field.z == 0.3
    assert decoded_msgs[0].magnetic_field_covariance == [0.01] * 9


def test_multi_dof_joint_state_pybag() -> None:
    transform = geometry_msgs.Transform(
        translation=geometry_msgs.Vector3(x=1.0, y=2.0, z=3.0),
        rotation=geometry_msgs.Quaternion(x=0.0, y=0.0, z=0.0, w=1.0)
    )
    twist = geometry_msgs.Twist(
        linear=geometry_msgs.Vector3(x=0.1, y=0.2, z=0.3),
        angular=geometry_msgs.Vector3(x=0.01, y=0.02, z=0.03)
    )
    wrench = geometry_msgs.Wrench(
        force=geometry_msgs.Vector3(x=10.0, y=20.0, z=30.0),
        torque=geometry_msgs.Vector3(x=1.0, y=2.0, z=3.0)
    )
    msg = sensor_msgs.MultiDOFJointState(
        header=_make_header(),
        joint_names=['joint1', 'joint2'],
        transforms=[transform, transform],
        twist=[twist, twist],
        wrench=[wrench, wrench]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'MultiDOFJointState'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].joint_names == ['joint1', 'joint2']
    assert len(decoded_msgs[0].transforms) == 2
    assert decoded_msgs[0].transforms[0].translation.x == 1.0
    assert decoded_msgs[0].transforms[0].translation.y == 2.0
    assert decoded_msgs[0].transforms[0].translation.z == 3.0
    assert decoded_msgs[0].transforms[0].rotation.x == 0.0
    assert decoded_msgs[0].transforms[0].rotation.y == 0.0
    assert decoded_msgs[0].transforms[0].rotation.z == 0.0
    assert decoded_msgs[0].transforms[0].rotation.w == 1.0
    assert decoded_msgs[0].transforms[1].translation.x == 1.0
    assert decoded_msgs[0].transforms[1].translation.y == 2.0
    assert decoded_msgs[0].transforms[1].translation.z == 3.0
    assert decoded_msgs[0].transforms[1].rotation.x == 0.0
    assert decoded_msgs[0].transforms[1].rotation.y == 0.0
    assert decoded_msgs[0].transforms[1].rotation.z == 0.0
    assert decoded_msgs[0].transforms[1].rotation.w == 1.0
    assert decoded_msgs[0].twist[0].angular.x == 0.01
    assert decoded_msgs[0].twist[0].angular.y == 0.02
    assert decoded_msgs[0].twist[0].angular.z == 0.03
    assert decoded_msgs[0].twist[1].angular.x == 0.01
    assert decoded_msgs[0].twist[1].angular.y == 0.02
    assert decoded_msgs[0].twist[1].angular.z == 0.03
    assert decoded_msgs[0].wrench[0].force.x == 10.0
    assert decoded_msgs[0].wrench[0].force.y == 20.0
    assert decoded_msgs[0].wrench[0].force.z == 30.0
    assert decoded_msgs[0].wrench[0].torque.x == 1.0
    assert decoded_msgs[0].wrench[0].torque.y == 2.0
    assert decoded_msgs[0].wrench[0].torque.z == 3.0
    assert decoded_msgs[0].wrench[1].force.x == 10.0
    assert decoded_msgs[0].wrench[1].force.y == 20.0
    assert decoded_msgs[0].wrench[1].force.z == 30.0
    assert decoded_msgs[0].wrench[1].torque.x == 1.0
    assert decoded_msgs[0].wrench[1].torque.y == 2.0
    assert decoded_msgs[0].wrench[1].torque.z == 3.0


def test_multi_echo_laser_scan_pybag() -> None:
    echo1 = sensor_msgs.LaserEcho(echoes=[1.0, 1.5])
    echo2 = sensor_msgs.LaserEcho(echoes=[2.0, 2.5])
    msg = sensor_msgs.MultiEchoLaserScan(
        header=_make_header(),
        angle_min=-1.50,
        angle_max=1.50,
        angle_increment=0.5,
        time_increment=0.25,
        scan_time=0.5,
        range_min=0.5,
        range_max=10.0,
        ranges=[echo1, echo2],
        intensities=[echo1, echo2]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'MultiEchoLaserScan'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].angle_min == -1.5
    assert decoded_msgs[0].angle_max == 1.5
    assert decoded_msgs[0].angle_increment == 0.5
    assert decoded_msgs[0].time_increment == 0.25
    assert decoded_msgs[0].scan_time == 0.5
    assert decoded_msgs[0].range_min == 0.5
    assert decoded_msgs[0].range_max == 10.0
    assert len(decoded_msgs[0].ranges) == 2
    assert decoded_msgs[0].ranges[0].echoes == [1.0, 1.5]
    assert decoded_msgs[0].ranges[1].echoes == [2.0, 2.5]


def test_nav_sat_status_pybag() -> None:
    msg = sensor_msgs.NavSatStatus(
        status=sensor_msgs.NavSatStatus.STATUS_FIX,
        service=sensor_msgs.NavSatStatus.SERVICE_GPS
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'NavSatStatus'
    assert decoded_msgs[0].status == 0   # STATUS_FIX
    assert decoded_msgs[0].service == 1  # SERVICE_GPS


def test_nav_sat_fix_pybag() -> None:
    status = sensor_msgs.NavSatStatus(
        status=sensor_msgs.NavSatStatus.STATUS_FIX,
        service=sensor_msgs.NavSatStatus.SERVICE_GPS
    )
    msg = sensor_msgs.NavSatFix(
        header=_make_header(),
        status=status,
        latitude=37.7749,
        longitude=-122.4194,
        altitude=100.0,
        position_covariance=[1.0] * 9,
        position_covariance_type=sensor_msgs.NavSatFix.COVARIANCE_TYPE_DIAGONAL_KNOWN
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'NavSatFix'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].status.status == 0  # STATUS_FIX
    assert decoded_msgs[0].status.service == 1  # SERVICE_GPS
    assert decoded_msgs[0].latitude == 37.7749
    assert decoded_msgs[0].longitude == -122.4194
    assert decoded_msgs[0].altitude == 100.0
    assert decoded_msgs[0].position_covariance == [1.0] * 9
    assert decoded_msgs[0].position_covariance_type == 2  # COVARIANCE_TYPE_DIAGONAL_KNOWN


def test_point_field_pybag() -> None:
    msg = sensor_msgs.PointField(
        name='x',
        offset=0,
        datatype=sensor_msgs.PointField.FLOAT32,
        count=1
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'PointField'
    assert decoded_msgs[0].name == 'x'
    assert decoded_msgs[0].offset == 0
    assert decoded_msgs[0].datatype == 7  # FLOAT32
    assert decoded_msgs[0].count == 1


def test_point_cloud_pybag() -> None:
    point1 = geometry_msgs.Point32(x=1.0, y=2.0, z=3.0)
    point2 = geometry_msgs.Point32(x=4.0, y=5.0, z=6.0)
    channel = sensor_msgs.ChannelFloat32(name='intensity', values=[100.0, 200.0])
    msg = sensor_msgs.PointCloud(
        header=_make_header(),
        points=[point1, point2],
        channels=[channel]
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'PointCloud'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert len(decoded_msgs[0].points) == 2
    assert decoded_msgs[0].points[0].x == 1.0
    assert decoded_msgs[0].points[0].y == 2.0
    assert decoded_msgs[0].points[0].z == 3.0
    assert decoded_msgs[0].points[1].x == 4.0
    assert decoded_msgs[0].points[1].y == 5.0
    assert decoded_msgs[0].points[1].z == 6.0
    assert len(decoded_msgs[0].channels) == 1
    assert decoded_msgs[0].channels[0].name == 'intensity'
    assert decoded_msgs[0].channels[0].values == [100.0, 200.0]


def test_point_cloud2_pybag() -> None:
    field1 = sensor_msgs.PointField(
        name='x',
        offset=0,
        datatype=sensor_msgs.PointField.FLOAT32,
        count=1
    )
    field2 = sensor_msgs.PointField(
        name='y',
        offset=4,
        datatype=sensor_msgs.PointField.FLOAT32,
        count=1
    )
    msg = sensor_msgs.PointCloud2(
        header=_make_header(),
        height=1,
        width=2,
        fields=[field1, field2],
        is_bigendian=False,
        point_step=8,
        row_step=16,
        data=[0, 0, 128, 63, 0, 0, 0, 64, 0, 0, 64, 64, 0, 0, 128, 64],  # Sample point data
        is_dense=True
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'PointCloud2'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].height == 1
    assert decoded_msgs[0].width == 2
    assert len(decoded_msgs[0].fields) == 2
    assert decoded_msgs[0].fields[0].name == 'x'
    assert decoded_msgs[0].fields[0].offset == 0
    assert decoded_msgs[0].fields[0].datatype == 7  # FLOAT32
    assert decoded_msgs[0].fields[0].count == 1
    assert decoded_msgs[0].fields[1].name == 'y'
    assert decoded_msgs[0].fields[1].offset == 4
    assert decoded_msgs[0].fields[1].datatype == 7  # FLOAT32
    assert decoded_msgs[0].fields[1].count == 1
    assert decoded_msgs[0].is_bigendian == False
    assert decoded_msgs[0].point_step == 8
    assert decoded_msgs[0].row_step == 16
    assert decoded_msgs[0].data == bytes([0, 0, 128, 63, 0, 0, 0, 64, 0, 0, 64, 64, 0, 0, 128, 64])
    assert decoded_msgs[0].is_dense == True


def test_range_pybag() -> None:
    msg = sensor_msgs.Range(
        header=_make_header(),
        radiation_type=sensor_msgs.Range.ULTRASOUND,
        field_of_view=0.5,
        min_range=0.25,
        max_range=10.0,
        range=2.5
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'Range'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].radiation_type == 0  # ULTRASOUND
    assert decoded_msgs[0].field_of_view == 0.5
    assert decoded_msgs[0].min_range == 0.25
    assert decoded_msgs[0].max_range == 10.0
    assert decoded_msgs[0].range == 2.5


def test_relative_humidity_pybag() -> None:
    msg = sensor_msgs.RelativeHumidity(
        header=_make_header(),
        relative_humidity=65.5,
        variance=0.1
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'RelativeHumidity'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].relative_humidity == 65.5
    assert decoded_msgs[0].variance == 0.1


def test_temperature_pybag() -> None:
    msg = sensor_msgs.Temperature(
        header=_make_header(),
        temperature=25.0,
        variance=0.1
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'Temperature'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].temperature == 25.0
    assert decoded_msgs[0].variance == 0.1


def test_time_reference_pybag() -> None:
    time_ref = builtin_interfaces.Time(sec=1000, nanosec=500000)
    msg = sensor_msgs.TimeReference(
        header=_make_header(),
        time_ref=time_ref,
        source='gps'
    )
    decoded_msgs = _roundtrip_write(msg)
    assert len(decoded_msgs) == 1
    assert decoded_msgs[0].__name__ == 'TimeReference'
    assert decoded_msgs[0].header.frame_id == 'frame'
    assert decoded_msgs[0].header.stamp.sec == 1
    assert decoded_msgs[0].header.stamp.nanosec == 2
    assert decoded_msgs[0].time_ref.sec == 1000
    assert decoded_msgs[0].time_ref.nanosec == 500000
    assert decoded_msgs[0].source == 'gps'



import logging
import struct
import zlib
from abc import ABC, abstractmethod
from typing import Any, Callable, Literal

from pybag.io.raw_writer import BaseWriter, BytesWriter, CrcWriter
from pybag.mcap.records import (
    AttachmentIndexRecord,
    AttachmentRecord,
    ChannelRecord,
    ChunkIndexRecord,
    ChunkRecord,
    DataEndRecord,
    FooterRecord,
    HeaderRecord,
    MessageIndexRecord,
    MessageRecord,
    MetadataIndexRecord,
    MetadataRecord,
    RecordType,
    SchemaRecord,
    StatisticsRecord,
    SummaryOffsetRecord
)

# Footer payload size: 8 bytes summary_start + 8 bytes summary_offset_start + 4 bytes summary_crc
FOOTER_PAYLOAD_SIZE = 20


class McapRecordWriter:
    """Utilities for writing MCAP records."""

    @classmethod
    def write_magic_bytes(cls, writer: BaseWriter, version: str = "0") -> int:
        """Write the MCAP magic bytes."""
        magic = b"\x89MCAP" + version.encode() + b"\r\n"
        return writer.write(magic)

    # Primitive encoders -------------------------------------------------

    @classmethod
    def _encode_uint8(cls, value: int) -> bytes:
        return struct.pack("<B", value)

    @classmethod
    def _encode_uint16(cls, value: int) -> bytes:
        return struct.pack("<H", value)

    @classmethod
    def _encode_uint32(cls, value: int) -> bytes:
        return struct.pack("<I", value)

    @classmethod
    def _encode_uint64(cls, value: int) -> bytes:
        return struct.pack("<Q", value)

    @classmethod
    def _encode_timestamp(cls, value: int) -> bytes:
        return cls._encode_uint64(value)

    @classmethod
    def _encode_record_type(cls, value: int) -> bytes:
        return cls._encode_uint8(value)

    @classmethod
    def _encode_string(cls, value: str) -> bytes:
        encoded = value.encode()
        return cls._encode_uint32(len(encoded)) + encoded

    # Container encoders -------------------------------------------------

    @classmethod
    def _encode_tuple(cls, data: tuple, first_type: str, second_type: str) -> bytes:
        first_value, second_value = data
        first_value_encoded = getattr(cls, f'_encode_{first_type}')(first_value)
        second_value_encoded = getattr(cls, f'_encode_{second_type}')(second_value)
        return first_value_encoded + second_value_encoded

    @classmethod
    def _encode_map(cls, data: dict, key_type: str, value_type: str) -> bytes:
        parts: list[bytes] = []
        for k, v in data.items():
            parts.append(getattr(cls, f'_encode_{key_type}')(k))
            parts.append(getattr(cls, f'_encode_{value_type}')(v))
        payload = b"".join(parts)
        return cls._encode_uint32(len(payload)) + payload

    @classmethod
    def _encode_array(cls, data: list, array_type_parser: Callable[[Any], bytes]) -> bytes:
        parts = [array_type_parser(v) for v in data]
        payload = b"".join(parts)
        return cls._encode_uint32(len(payload)) + payload

    # MCAP Record Writers -----------------------------------------------

    @classmethod
    def _write_record(cls, writer: BaseWriter, record_type: int, payload: bytes) -> None:
        writer.write(cls._encode_record_type(record_type))
        writer.write(cls._encode_uint64(len(payload)))
        writer.write(payload)

    @classmethod
    def write_header(cls, writer: BaseWriter, record: HeaderRecord) -> None:
        payload = (
              cls._encode_string(record.profile)
            + cls._encode_string(record.library)
        )
        cls._write_record(writer, RecordType.HEADER, payload)


    @classmethod
    def write_footer(cls, writer: BaseWriter, record: FooterRecord) -> None:
        payload = (
              cls._encode_uint64(record.summary_start)
            + cls._encode_uint64(record.summary_offset_start)
            + cls._encode_uint32(record.summary_crc)
        )
        cls._write_record(writer, RecordType.FOOTER, payload)

    @classmethod
    def write_schema(cls, writer: BaseWriter, record: SchemaRecord) -> None:
        payload = (
              cls._encode_uint16(record.id)
            + cls._encode_string(record.name)
            + cls._encode_string(record.encoding)
            + cls._encode_uint32(len(record.data))
            + record.data  # just bytes
        )
        cls._write_record(writer, RecordType.SCHEMA, payload)

    @classmethod
    def write_channel(cls, writer: BaseWriter, record: ChannelRecord) -> None:
        payload = (
              cls._encode_uint16(record.id)
            + cls._encode_uint16(record.schema_id)
            + cls._encode_string(record.topic)
            + cls._encode_string(record.message_encoding)
            + cls._encode_map(record.metadata, "string", "string")
        )
        cls._write_record(writer, RecordType.CHANNEL, payload)


    @classmethod
    def write_message(cls, writer: BaseWriter, record: MessageRecord) -> None:
        payload = (
              cls._encode_uint16(record.channel_id)
            + cls._encode_uint32(record.sequence)
            + cls._encode_timestamp(record.log_time)
            + cls._encode_timestamp(record.publish_time)
            + record.data
        )
        cls._write_record(writer, RecordType.MESSAGE, payload)

    @classmethod
    def write_chunk(cls, writer: BaseWriter, record: ChunkRecord) -> None:
        payload = (
              cls._encode_timestamp(record.message_start_time)
            + cls._encode_timestamp(record.message_end_time)
            + cls._encode_uint64(record.uncompressed_size)
            + cls._encode_uint32(record.uncompressed_crc)
            + cls._encode_string(record.compression)
            + cls._encode_uint64(len(record.records))
            + record.records
        )
        cls._write_record(writer, RecordType.CHUNK, payload)


    @classmethod
    def write_message_index(cls, writer: BaseWriter, record: MessageIndexRecord) -> None:
        payload = (
              cls._encode_uint16(record.channel_id)
            + cls._encode_array(record.records, lambda x: cls._encode_tuple(x, "timestamp", "uint64"))
        )
        cls._write_record(writer, RecordType.MESSAGE_INDEX, payload)

    @classmethod
    def write_chunk_index(cls, writer: BaseWriter, record: ChunkIndexRecord) -> None:
        payload = (
              cls._encode_timestamp(record.message_start_time)
            + cls._encode_timestamp(record.message_end_time)
            + cls._encode_uint64(record.chunk_start_offset)
            + cls._encode_uint64(record.chunk_length)
            + cls._encode_map(record.message_index_offsets, "uint16", "uint64")
            + cls._encode_uint64(record.message_index_length)
            + cls._encode_string(record.compression)
            + cls._encode_uint64(record.compressed_size)
            + cls._encode_uint64(record.uncompressed_size)
        )
        cls._write_record(writer, RecordType.CHUNK_INDEX, payload)

    @classmethod
    def write_attachment(cls, writer: BaseWriter, record: AttachmentRecord) -> None:
        payload = (
              cls._encode_timestamp(record.log_time)
            + cls._encode_timestamp(record.create_time)
            + cls._encode_string(record.name)
            + cls._encode_string(record.media_type)
            + cls._encode_uint64(len(record.data))
            + record.data
            + cls._encode_uint32(record.crc)
        )
        cls._write_record(writer, RecordType.ATTACHMENT, payload)

    @classmethod
    def write_metadata(cls, writer: BaseWriter, record: MetadataRecord) -> None:
        payload = (
              cls._encode_string(record.name)
            + cls._encode_map(record.metadata, "string", "string")
        )
        cls._write_record(writer, RecordType.METADATA, payload)

    @classmethod
    def write_data_end(cls, writer: BaseWriter, record: DataEndRecord) -> None:
        payload = cls._encode_uint32(record.data_section_crc)
        cls._write_record(writer, RecordType.DATA_END, payload)

    @classmethod
    def write_attachment_index(cls, writer: BaseWriter, record: AttachmentIndexRecord) -> None:
        payload = (
              cls._encode_uint64(record.offset)
            + cls._encode_uint64(record.length)
            + cls._encode_timestamp(record.log_time)
            + cls._encode_timestamp(record.create_time)
            + cls._encode_uint64(record.data_size)
            + cls._encode_string(record.name)
            + cls._encode_string(record.media_type)
        )
        cls._write_record(writer, RecordType.ATTACHMENT_INDEX, payload)

    @classmethod
    def write_metadata_index(cls, writer: BaseWriter, record: MetadataIndexRecord) -> None:
        payload = (
              cls._encode_uint64(record.offset)
            + cls._encode_uint64(record.length)
            + cls._encode_string(record.name)
        )
        cls._write_record(writer, RecordType.METADATA_INDEX, payload)

    @classmethod
    def write_statistics(cls, writer: BaseWriter, record: StatisticsRecord) -> None:
        payload = (
              cls._encode_uint64(record.message_count)
            + cls._encode_uint16(record.schema_count)
            + cls._encode_uint32(record.channel_count)
            + cls._encode_uint32(record.attachment_count)
            + cls._encode_uint32(record.metadata_count)
            + cls._encode_uint32(record.chunk_count)
            + cls._encode_timestamp(record.message_start_time)
            + cls._encode_timestamp(record.message_end_time)
            + cls._encode_map(record.channel_message_counts, "uint16", "uint64")
        )
        cls._write_record(writer, RecordType.STATISTICS, payload)

    @classmethod
    def write_summary_offset(cls, writer: BaseWriter, record: SummaryOffsetRecord) -> None:
        payload = (
              cls._encode_uint8(record.group_opcode)
            + cls._encode_uint64(record.group_start)
            + cls._encode_uint64(record.group_length)
        )
        cls._write_record(writer, RecordType.SUMMARY_OFFSET, payload)


# Low-level MCAP Writer Classes ------------------------------------------


class BaseMcapRecordWriter(ABC):
    """Abstract base class for low-level MCAP record writers.

    Low-level writers accept pre-constructed MCAP record dataclasses and handle
    binary serialization, chunking, compression, and summary section management.
    """

    @abstractmethod
    def write_schema(self, schema: SchemaRecord) -> None:
        """Write a schema record to the MCAP file.

        Args:
            schema: The schema record to write.
        """
        ...  # pragma: no cover

    @abstractmethod
    def write_channel(self, channel: ChannelRecord) -> None:
        """Write a channel record to the MCAP file.

        Args:
            channel: The channel record to write.
        """
        ...  # pragma: no cover

    @abstractmethod
    def write_message(self, message: MessageRecord) -> None:
        """Write a message record to the MCAP file.

        Args:
            message: The message record to write.
        """
        ...  # pragma: no cover

    @abstractmethod
    def close(self) -> None:
        """Finalize the MCAP file by writing summary section and footer."""
        ...  # pragma: no cover

    @abstractmethod
    def __enter__(self) -> 'BaseMcapRecordWriter':
        """Context manager entry."""
        ...  # pragma: no cover

    @abstractmethod
    def __exit__(self, exc_type, exc, tb) -> None:
        """Context manager exit."""
        ...  # pragma: no cover

    def _write_summary_section(
        self,
        writer: CrcWriter,
        schema_records: list[SchemaRecord] | None = None,
        channel_records: list[ChannelRecord] | None = None,
        statistics_record: StatisticsRecord | None = None,
        chunk_indexes: list[ChunkIndexRecord] | None = None
    ) -> tuple[int, int]:
        """Write the summary section and return (summary_start, summary_offset_start).

        Args:
            chunk_indexes: Optional list of chunk index records (for chunked writers).

        Returns:
            Tuple of (summary_start, summary_offset_start) positions.
        """
        # Start summary section
        summary_start = writer.tell()
        writer.clear_crc()

        # Write schema records to summary
        schema_group_start = summary_start
        if schema_records:
            for record in schema_records:
                McapRecordWriter.write_schema(writer, record)
        schema_group_length = writer.tell() - schema_group_start

        # Write channel records to summary
        channel_group_start = writer.tell()
        if channel_records:
            for record in channel_records:
                McapRecordWriter.write_channel(writer, record)
        channel_group_length = writer.tell() - channel_group_start

        # Write chunk index records to summary (only for chunked writers)
        chunk_index_group_start = writer.tell()
        chunk_index_group_length = 0
        if chunk_indexes:
            for record in chunk_indexes:
                McapRecordWriter.write_chunk_index(writer, record)
            chunk_index_group_length = writer.tell() - chunk_index_group_start

        # Write statistics record
        statistics_group_start = writer.tell()
        if statistics_record is not None:
            McapRecordWriter.write_statistics(writer, statistics_record)
        statistics_group_length = writer.tell() - statistics_group_start

        # Write summary offsets
        summary_offset_start = writer.tell()
        if schema_group_length > 0:
            McapRecordWriter.write_summary_offset(
                writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.SCHEMA,
                    group_start=schema_group_start,
                    group_length=schema_group_length,
                ),
            )
        if channel_group_length > 0:
            McapRecordWriter.write_summary_offset(
                writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.CHANNEL,
                    group_start=channel_group_start,
                    group_length=channel_group_length,
                ),
            )
        if chunk_index_group_length > 0:
            McapRecordWriter.write_summary_offset(
                writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.CHUNK_INDEX,
                    group_start=chunk_index_group_start,
                    group_length=chunk_index_group_length,
                ),
            )
        if statistics_group_length > 0:
            McapRecordWriter.write_summary_offset(
                writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.STATISTICS,
                    group_start=statistics_group_start,
                    group_length=statistics_group_length,
                ),
            )

        return summary_start, summary_offset_start


class McapNonChunkedWriter(BaseMcapRecordWriter):
    """Low-level MCAP writer for non-chunked files.

    Writes schemas, channels, and messages directly to the data section without
    chunking. Tracks all records and statistics for the summary section.
    """

    def __init__(self, writer: BaseWriter, *, profile: str = "ros2") -> None:
        """Initialize a non-chunked MCAP writer.

        Args:
            writer: The underlying writer to write binary data to.
            profile: The MCAP profile to use (default: "ros2").
        """
        self._writer = CrcWriter(writer)
        self._profile = profile

        # Tracking for summary section
        self._schema_records: dict[int, SchemaRecord] = {}
        self._channel_records: dict[int, ChannelRecord] = {}

        # Statistics tracking
        self._message_count = 0
        self._chunk_count = 0
        self._message_start_time: int | None = None
        self._message_end_time: int | None = None
        self._channel_message_counts: dict[int, int] = {}

        # Write file header
        from pybag import __version__
        McapRecordWriter.write_magic_bytes(self._writer)
        header = HeaderRecord(profile=profile, library=f"pybag {__version__}")
        McapRecordWriter.write_header(self._writer, header)

    def __enter__(self) -> 'McapNonChunkedWriter':
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    def write_schema(self, schema: SchemaRecord) -> None:
        """Write a schema record immediately to the data section."""
        McapRecordWriter.write_schema(self._writer, schema)
        if schema.id in self._schema_records:
            logging.warning(f'Schema (id {schema.id}) already written to file')
        else:
            self._schema_records[schema.id] = schema

    def write_channel(self, channel: ChannelRecord) -> None:
        """Write a channel record immediately to the data section."""
        McapRecordWriter.write_channel(self._writer, channel)
        if channel.id in self._channel_records:
            logging.warning(f'Channel (id {channel.id}) already written to file')
        else:
            self._channel_records[channel.id] = channel
            self._channel_message_counts[channel.id] = 0

    def write_message(self, message: MessageRecord) -> None:
        """Write a message record immediately to the data section."""
        McapRecordWriter.write_message(self._writer, message)

        # Update statistics
        self._message_count += 1
        self._channel_message_counts[message.channel_id] += 1
        self._message_start_time = min(
            self._message_start_time or message.log_time,
            message.log_time
        )
        self._message_end_time = max(
            self._message_end_time or message.log_time,
            message.log_time
        )

    def close(self) -> None:
        """Finalize the file by writing summary section and footer."""
        # Write DataEnd record
        data_end = DataEndRecord(data_section_crc=self._writer.get_crc())
        McapRecordWriter.write_data_end(self._writer, data_end)

        # Write summary section using shared helper
        summary_start, summary_offset_start = self._write_summary_section(
            self._writer,
            schema_records=list(self._schema_records.values()),
            channel_records=list(self._channel_records.values()),
            statistics_record=StatisticsRecord(
                message_count=self._message_count,
                schema_count=len(self._schema_records),
                channel_count=len(self._channel_records),
                attachment_count=0,  # TODO: Implement
                metadata_count=0,    # TODO: Implement
                chunk_count=self._chunk_count,
                message_start_time=self._message_start_time or 0,
                message_end_time=self._message_end_time or 0,
                channel_message_counts=self._channel_message_counts,
            )
        )

        # Write footer record manually for CRC calculation
        self._writer.write(McapRecordWriter._encode_record_type(RecordType.FOOTER))
        self._writer.write(McapRecordWriter._encode_uint64(FOOTER_PAYLOAD_SIZE))
        self._writer.write(McapRecordWriter._encode_uint64(summary_start))
        self._writer.write(McapRecordWriter._encode_uint64(summary_offset_start))
        self._writer.write(McapRecordWriter._encode_uint32(self._writer.get_crc()))

        # Write magic bytes again
        McapRecordWriter.write_magic_bytes(self._writer)

        # Close the underlying writer
        self._writer.close()


class McapChunkedWriter(BaseMcapRecordWriter):
    """Low-level MCAP writer for chunked files with compression.

    Writes schemas and channels directly to the file, but accumulates messages
    in chunks. When a chunk reaches the size threshold, it is compressed and
    written along with message indexes.
    """

    def __init__(
        self,
        writer: BaseWriter,
        *,
        chunk_size: int,
        chunk_compression: Literal["lz4", "zstd"] | None = None,
        profile: str = "ros2",
    ) -> None:
        """Initialize a chunked MCAP writer.

        Args:
            writer: The underlying writer to write binary data to.
            chunk_size: The size threshold for flushing chunks (in bytes).
            chunk_compression: Compression algorithm ("lz4" or "zstd").
            profile: The MCAP profile to use (default: "ros2").
        """
        self._writer = CrcWriter(writer)
        self._profile = profile
        self._chunk_size = chunk_size
        self._chunk_compression = chunk_compression or ""
        self._compress_chunk = self._create_chunk_compressor()

        # Tracking for summary section
        self._schema_records: dict[int, SchemaRecord] = {}
        self._channel_records: dict[ int, ChannelRecord] = {}
        self._chunk_indexes: list[ChunkIndexRecord] = []

        # Statistics tracking
        self._message_count = 0
        self._chunk_count = 0
        self._message_start_time: int | None = None
        self._message_end_time: int | None = None
        self._channel_message_counts: dict[int, int] = {}

        # Current chunk buffering
        self._current_chunk_buffer: BytesWriter = BytesWriter()
        self._current_chunk_start_time: int | None = None
        self._current_chunk_end_time: int | None = None
        self._current_message_index: dict[int, list[tuple[int, int]]] = {}

        # Write file header
        from pybag import __version__
        McapRecordWriter.write_magic_bytes(self._writer)
        header = HeaderRecord(profile=profile, library=f"pybag {__version__}")
        McapRecordWriter.write_header(self._writer, header)

    def __enter__(self) -> 'McapChunkedWriter':
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    def _create_chunk_compressor(self) -> Callable[[bytes], bytes]:
        """Create a compression function based on the configured algorithm."""
        if self._chunk_compression == "lz4":
            import lz4.frame
            return lz4.frame.compress
        elif self._chunk_compression == "zstd":
            import zstandard as zstd
            return zstd.ZstdCompressor().compress
        elif self._chunk_compression == "":
            return lambda x: x
        else:
            raise ValueError(f"Unsupported chunk compression: {self._chunk_compression}")

    def write_schema(self, schema: SchemaRecord) -> None:
        """Write a schema record immediately to the data section (not buffered)."""
        McapRecordWriter.write_schema(self._writer, schema)
        if schema.id in self._schema_records:
            logging.warning(f'Schema (id {schema.id}) already written to file')
        else:
            self._schema_records[schema.id] = schema

    def write_channel(self, channel: ChannelRecord) -> None:
        """Write a channel record immediately to the data section (not buffered)."""
        McapRecordWriter.write_channel(self._writer, channel)
        if channel.id in self._channel_records:
            logging.warning(f'Channel (id {channel.id}) already written to file')
        else:
            self._channel_records[channel.id] = channel
            self._channel_message_counts[channel.id] = 0

    def write_message(self, message: MessageRecord) -> None:
        """Write a message record to the current chunk buffer.

        If the buffer size exceeds the chunk size threshold, flush the chunk.
        """
        # Update chunk timing
        self._current_chunk_start_time = min(
            self._current_chunk_start_time or message.log_time,
            message.log_time
        )
        self._current_chunk_end_time = max(
            self._current_chunk_end_time or message.log_time,
            message.log_time
        )

        # Write message to chunk buffer and track offset
        offset = self._current_chunk_buffer.size()
        McapRecordWriter.write_message(self._current_chunk_buffer, message)
        self._current_message_index.setdefault(message.channel_id, []).append(
            (message.log_time, offset)
        )

        # Flush chunk if size threshold reached
        if self._current_chunk_buffer.size() >= self._chunk_size:
            self._flush_chunk()

        # Update statistics
        self._message_count += 1
        self._channel_message_counts[message.channel_id] += 1
        self._message_start_time = min(
            self._message_start_time or message.log_time,
            message.log_time
        )
        self._message_end_time = max(
            self._message_end_time or message.log_time,
            message.log_time
        )

    def _flush_chunk(self) -> None:
        """Compress and write the current chunk buffer to the file."""
        records = self._current_chunk_buffer.as_bytes()

        # Create and write chunk record
        chunk = ChunkRecord(
            message_start_time=self._current_chunk_start_time or 0,
            message_end_time=self._current_chunk_end_time or 0,
            uncompressed_size=len(records),
            uncompressed_crc=zlib.crc32(records),
            compression=self._chunk_compression,
            records=self._compress_chunk(records),
        )
        chunk_start_offset = self._writer.tell()
        McapRecordWriter.write_chunk(self._writer, chunk)
        chunk_length = self._writer.tell() - chunk_start_offset

        # Write message index records for the chunk
        message_index_offsets = {}
        message_index_start_offset = self._writer.tell()
        for cid, records_list in self._current_message_index.items():
            message_index_offsets[cid] = self._writer.tell()
            message_index_record = MessageIndexRecord(
                channel_id=cid,
                records=records_list
            )
            McapRecordWriter.write_message_index(self._writer, message_index_record)
        message_index_length = self._writer.tell() - message_index_start_offset

        # Track chunk index for summary
        self._chunk_indexes.append(
            ChunkIndexRecord(
                message_start_time=chunk.message_start_time,
                message_end_time=chunk.message_end_time,
                chunk_start_offset=chunk_start_offset,
                chunk_length=chunk_length,
                message_index_offsets=message_index_offsets,
                message_index_length=message_index_length,
                compression=chunk.compression,
                compressed_size=len(chunk.records),
                uncompressed_size=chunk.uncompressed_size,
            )
        )

        # Increment chunk count and clear buffer
        self._chunk_count += 1
        self._current_chunk_buffer.clear()
        self._current_chunk_start_time = None
        self._current_chunk_end_time = None
        self._current_message_index = {}

    def close(self) -> None:
        """Finalize the file by flushing remaining chunk and writing summary."""
        # Flush any remaining buffered messages
        if self._current_chunk_buffer.size() > 0:
            self._flush_chunk()

        # Write DataEnd record
        data_end = DataEndRecord(data_section_crc=self._writer.get_crc())
        McapRecordWriter.write_data_end(self._writer, data_end)

        # Write summary section using shared helper (passing chunk indexes)
        summary_start, summary_offset_start = self._write_summary_section(
            self._writer,
            schema_records=list(self._schema_records.values()),
            channel_records=list(self._channel_records.values()),
            statistics_record=StatisticsRecord(
                message_count=self._message_count,
                schema_count=len(self._schema_records),
                channel_count=len(self._channel_records),
                attachment_count=0,  # TODO: Implement
                metadata_count=0,    # TODO: Implement
                chunk_count=self._chunk_count,
                message_start_time=self._message_start_time or 0,
                message_end_time=self._message_end_time or 0,
                channel_message_counts=self._channel_message_counts,
            ),
            chunk_indexes=self._chunk_indexes
        )

        # Write footer record manually for CRC calculation
        self._writer.write(McapRecordWriter._encode_record_type(RecordType.FOOTER))
        self._writer.write(McapRecordWriter._encode_uint64(FOOTER_PAYLOAD_SIZE))
        self._writer.write(McapRecordWriter._encode_uint64(summary_start))
        self._writer.write(McapRecordWriter._encode_uint64(summary_offset_start))
        self._writer.write(McapRecordWriter._encode_uint32(self._writer.get_crc()))

        # Write magic bytes again
        McapRecordWriter.write_magic_bytes(self._writer)

        # Close the underlying writer
        self._writer.close()


class McapRecordWriterFactory:
    """Factory for creating appropriate MCAP record writers."""

    @staticmethod
    def create_writer(
        writer: BaseWriter,
        *,
        chunk_size: int | None = None,
        chunk_compression: Literal["lz4", "zstd"] | None = None,
        profile: str = "ros2",
    ) -> BaseMcapRecordWriter:
        """Create an appropriate MCAP record writer based on configuration.

        Args:
            writer: The underlying writer to write binary data to.
            chunk_size: If provided, creates a chunked writer with this size threshold.
                       If None, creates a non-chunked writer.
            chunk_compression: Compression algorithm for chunks ("lz4" or "zstd").
            profile: The MCAP profile to use (default: "ros2").

        Returns:
            A BaseMcapRecordWriter instance (either chunked or non-chunked).
        """
        if chunk_size is None:
            return McapNonChunkedWriter(writer, profile=profile)
        else:
            return McapChunkedWriter(
                writer,
                chunk_size=chunk_size,
                chunk_compression=chunk_compression,
                profile=profile,
            )



import logging
import struct
from enum import IntEnum
from typing import Any, Callable, Iterator

from pybag.io.raw_reader import BaseReader, BytesReader
from pybag.mcap.records import (
    AttachmentIndexRecord,
    AttachmentRecord,
    ChannelRecord,
    ChunkIndexRecord,
    ChunkRecord,
    DataEndRecord,
    FooterRecord,
    HeaderRecord,
    MessageIndexRecord,
    MessageRecord,
    MetadataIndexRecord,
    MetadataRecord,
    RecordType,
    SchemaRecord,
    StatisticsRecord,
    SummaryOffsetRecord
)

logger = logging.getLogger(__name__)

# Number of bytes of fixed-sized records
MAGIC_BYTES_SIZE = 8
FOOTER_SIZE = 29  # Includes the 1 byte record type and 8 bytes record length
DATA_END_SIZE = 13  # Includes the 1 byte record type and 8 bytes record length

class McapRecordType(IntEnum):
    HEADER = 1
    FOOTER = 2
    SCHEMA = 3
    CHANNEL = 4
    MESSAGE = 5
    CHUNK = 6
    MESSAGE_INDEX = 7
    CHUNK_INDEX = 8
    ATTACHMENT = 9
    ATTACHMENT_INDEX = 10
    STATISTICS = 11
    METADATA = 12
    METADATA_INDEX = 13
    SUMMARY_OFFSET = 14
    DATA_END = 15


class MalformedMCAP(Exception):
    """The MCAP format does not conform to the specification."""
    def __init__(self, error_msessage: str):
        super().__init__(error_msessage)


class McapRecordParser:
    @classmethod
    def peek_record(cls, file: BaseReader) -> int:
        """Peek at the next record in the MCAP file."""
        # If peek(1) returns b'', then this returns 0
        # No record has an ID of 0 so this indicates end
        return int.from_bytes(file.peek(1)[:1], 'little')


    @classmethod
    def skip_record(cls, file: BaseReader) -> None:
        """Skip the next record in the MCAP file."""
        _ = file.read(1)  # Skip the record type
        _, record_length = cls._parse_uint64(file)
        file.seek_from_current(record_length)


    @classmethod
    def parse_record(cls, file: BaseReader) -> Iterator[tuple[int, Any]]:
        """Read the next record in the MCAP file."""
        while True:
            record_type = cls.peek_record(file)
            logger.debug(f'Peeked at {record_type} record...')
            if record_type == 0:
                break  # EOF
            yield record_type, cls._parse_record(record_type, file)


    @classmethod
    def parse_magic_bytes(cls, file: BaseReader) -> str:
        """Parse the magic bytes at the begining/end of the MCAP file."""
        magic = file.read(8)
        if magic != b'\x89MCAP\x30\r\n':  # TODO: Support multiple versions
            raise MalformedMCAP(f'Invalid magic bytes: {str(magic)}')
        return chr(magic[5])  # Return the version


    @classmethod
    def _parse_record(cls, record_type: int, file: BaseReader) -> Any:
        """Parse the next record in the MCAP file."""
        record_name = RecordType(record_type).name.lower()
        logger.debug(f'Parsing {record_name} record...')
        return getattr(cls, f'parse_{record_name}')(file)

    # MCAP Serialization Handlers

    @classmethod
    def _parse_uint8(cls, file: BaseReader) -> tuple[int, int]:
        return 1, struct.unpack('<B', file.read(1))[0]


    @classmethod
    def _parse_uint16(cls, file: BaseReader) -> tuple[int, int]:
        return 2, struct.unpack('<H', file.read(2))[0]


    @classmethod
    def _parse_uint32(cls, file: BaseReader) -> tuple[int, int]:
        return 4, struct.unpack('<I', file.read(4))[0]


    @classmethod
    def _parse_uint64(cls, file: BaseReader) -> tuple[int, int]:
        return 8, struct.unpack('<Q', file.read(8))[0]


    @classmethod
    def _parse_string(cls, file: BaseReader) -> tuple[int, str]:
        string_length_bytes, string_length = cls._parse_uint32(file)
        string = file.read(string_length)
        return string_length_bytes + string_length, string.decode()


    @classmethod
    def _parse_timestamp(cls, file: BaseReader) -> tuple[int, int]:
        return cls._parse_uint64(file)


    @classmethod
    def _parse_bytes(cls, file: BaseReader, size: int) -> tuple[int, bytes]:
        bytes = file.read(size)
        return len(bytes), bytes


    @classmethod
    def _parse_tuple(cls, file: BaseReader, first_type: str, second_type: str) -> tuple[int, tuple]:
        first_value_length, first_value = getattr(cls, f'_parse_{first_type}')(file)
        second_value_length, second_value = getattr(cls, f'_parse_{second_type}')(file)
        return first_value_length + second_value_length, (first_value, second_value)


    @classmethod
    def _parse_map(cls, file: BaseReader, key_type: str, value_type: str) -> tuple[int, dict]:
        map_length_bytes, map_length = cls._parse_uint32(file)
        original_length = map_length

        map_key_value = {}
        while map_length > 0:
            key_length, key = getattr(cls, f'_parse_{key_type}')(file)
            value_length, value = getattr(cls, f'_parse_{value_type}')(file)
            map_key_value[key] = value
            map_length -= key_length + value_length

        if map_length != 0:
            bytes_read = original_length - map_length
            raise MalformedMCAP(f'Read {bytes_read} bytes. Expected {original_length}.')

        return map_length_bytes + map_length, map_key_value


    @classmethod
    def _parse_array(
        cls,
        file: BaseReader,
        array_type_parser: Callable[[BaseReader], tuple[int, Any]]
    ) -> tuple[int, list]:
        array_length_bytes, array_length = cls._parse_uint32(file)
        original_length = array_length

        array = []
        while array_length > 0:
            value_length, value = array_type_parser(file)
            array.append(value)
            array_length -= value_length

        if array_length != 0:
            bytes_read = original_length - array_length
            raise MalformedMCAP(f'Read {bytes_read} bytes. Expected {original_length}.')

        return array_length_bytes + array_length, array

    # MCAP Record Handlers

    @classmethod
    def parse_header(cls, file: BaseReader) -> HeaderRecord:
        """Parse the header record of an MCAP file."""
        if (record_type := file.read(1)) != b'\x01':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, profile = cls._parse_string(bytes_reader)
        _, library = cls._parse_string(bytes_reader)

        return HeaderRecord(profile, library)


    @classmethod
    def parse_footer(cls, file: BaseReader) -> FooterRecord:
        """Parse the footer record of an MCAP file."""
        if (record_type := file.read(1)) != b'\x02':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        # Footer record length is fixed to 20 bytes
        _, record_length = cls._parse_uint64(file)
        if record_length != 20:
            raise MalformedMCAP(f'Unexpected footer record length ({record_length} bytes).')
        bytes_reader = BytesReader(file.read(record_length))

        _, summary_start = cls._parse_uint64(bytes_reader)
        _, summary_offset_start = cls._parse_uint64(bytes_reader)
        _, summary_crc = cls._parse_uint32(bytes_reader)

        return FooterRecord(summary_start, summary_offset_start, summary_crc)


    @classmethod
    def parse_schema(cls, file: BaseReader) -> SchemaRecord | None:
        if (record_type := file.read(1)) != b'\x03':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, id = cls._parse_uint16(bytes_reader)
        if id == 0:  # Invalid and should be ignored
            return None

        _, name = cls._parse_string(bytes_reader)
        _, encoding = cls._parse_string(bytes_reader)
        _, data_length = cls._parse_uint32(bytes_reader)
        _, data = cls._parse_bytes(bytes_reader, data_length)

        return SchemaRecord(id, name, encoding, data)


    @classmethod
    def parse_channel(cls, file: BaseReader) -> ChannelRecord:
        if (record_type := file.read(1)) != b'\x04':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, id = cls._parse_uint16(bytes_reader)
        _, channel_id = cls._parse_uint16(bytes_reader)
        _, topic = cls._parse_string(bytes_reader)
        _, message_encoding = cls._parse_string(bytes_reader)
        _, metadata = cls._parse_map(bytes_reader, "string", "string")

        return ChannelRecord(id, channel_id, topic, message_encoding, metadata)


    @classmethod
    def parse_message(cls, file: BaseReader) -> MessageRecord:
        if (record_type := file.read(1)) != b'\x05':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, channel_id = cls._parse_uint16(bytes_reader)
        _, sequence = cls._parse_uint32(bytes_reader)
        _, log_time = cls._parse_timestamp(bytes_reader)
        _, publish_time = cls._parse_timestamp(bytes_reader)
        # Other fields: 2 + 4 + 8 + 8 = 22 bytes
        _, data = cls._parse_bytes(bytes_reader, record_length - 22)

        return MessageRecord(channel_id, sequence, log_time, publish_time, data)


    @classmethod
    def parse_chunk(cls, file: BaseReader) -> ChunkRecord:
        if (record_type := file.read(1)) != b'\x06':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, message_start_time = cls._parse_timestamp(bytes_reader)
        _, message_end_time = cls._parse_timestamp(bytes_reader)
        _, uncompressed_size = cls._parse_uint64(bytes_reader)
        _, uncompressed_crc = cls._parse_uint32(bytes_reader)
        _, compression = cls._parse_string(bytes_reader)
        _, records_length = cls._parse_uint64(bytes_reader)
        _, records = cls._parse_bytes(bytes_reader, records_length)

        return ChunkRecord(
            message_start_time,
            message_end_time,
            uncompressed_size,
            uncompressed_crc,
            compression,
            records
        )


    @classmethod
    def parse_message_index(cls, file: BaseReader) -> MessageIndexRecord:
        if (record_type := file.read(1)) != b'\x07':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, message_index_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(message_index_length))

        _, channel_id = cls._parse_uint16(bytes_reader)
        _, records = cls._parse_array(bytes_reader, lambda file: cls._parse_tuple(file, "timestamp", "uint64"))

        return MessageIndexRecord(channel_id, records)


    @classmethod
    def parse_chunk_index(cls, file: BaseReader) -> ChunkIndexRecord:
        if (record_type := file.read(1)) != b'\x08':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, message_start_time = cls._parse_timestamp(bytes_reader)
        _, message_end_time = cls._parse_timestamp(bytes_reader)
        _, chunk_start_offset = cls._parse_uint64(bytes_reader)
        _, chunk_length = cls._parse_uint64(bytes_reader)
        _, message_index_offsets = cls._parse_map(bytes_reader, "uint16", "uint64")
        _, message_index_length = cls._parse_uint64(bytes_reader)
        _, compression = cls._parse_string(bytes_reader)
        _, compressed_size = cls._parse_uint64(bytes_reader)
        _, uncompressed_size = cls._parse_uint64(bytes_reader)

        return ChunkIndexRecord(
            message_start_time,
            message_end_time,
            chunk_start_offset,
            chunk_length,
            message_index_offsets,
            message_index_length,
            compression,
            compressed_size,
            uncompressed_size
        )


    @classmethod
    def parse_attachment(cls, file: BaseReader) -> AttachmentRecord:
        if (record_type := file.read(1)) != b'\x09':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, log_time = cls._parse_timestamp(bytes_reader)
        _, create_time = cls._parse_timestamp(bytes_reader)
        _, name = cls._parse_string(bytes_reader)
        _, media_type = cls._parse_string(bytes_reader)
        _, data_bytes_length = cls._parse_uint64(bytes_reader)
        _, data_bytes = cls._parse_bytes(bytes_reader, data_bytes_length)
        _, crc = cls._parse_uint32(bytes_reader)

        return AttachmentRecord(log_time, create_time, name, media_type, data_bytes, crc)


    @classmethod
    def parse_metadata(cls, file: BaseReader) -> MetadataRecord:
        if (record_type := file.read(1)) != b'\x0C':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, name = cls._parse_string(bytes_reader)
        logger.debug(f'Parsing metadata for {name}...')
        _, metadata = cls._parse_map(bytes_reader, "string", "string")

        return MetadataRecord(name, metadata)


    @classmethod
    def parse_data_end(cls, file: BaseReader) -> DataEndRecord:
        if (record_type := file.read(1)) != b'\x0f':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, data_section_crc = cls._parse_uint32(bytes_reader)
        return DataEndRecord(data_section_crc)


    @classmethod
    def parse_attachment_index(cls, file: BaseReader) -> AttachmentIndexRecord:
        if (record_type := file.read(1)) != b'\x0A':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, offset = cls._parse_uint64(bytes_reader)
        _, length = cls._parse_uint64(bytes_reader)
        _, log_time = cls._parse_timestamp(bytes_reader)
        _, create_time = cls._parse_timestamp(bytes_reader)
        _, data_size = cls._parse_uint64(bytes_reader)
        _, name = cls._parse_string(bytes_reader)
        _, media_type = cls._parse_string(bytes_reader)

        return AttachmentIndexRecord(
            offset,
            length,
            log_time,
            create_time,
            data_size,
            name,
            media_type
        )


    @classmethod
    def parse_metadata_index(cls, file: BaseReader) -> MetadataIndexRecord:
        if (record_type := file.read(1)) != b'\x0D':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, offset = cls._parse_uint64(bytes_reader)
        _, length = cls._parse_uint64(bytes_reader)
        _, name = cls._parse_string(bytes_reader)

        return MetadataIndexRecord(offset, length, name)


    @classmethod
    def parse_statistics(cls, file: BaseReader) -> StatisticsRecord:
        if (record_type := file.read(1)) != b'\x0B':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, message_count = cls._parse_uint64(bytes_reader)
        _, schema_count = cls._parse_uint16(bytes_reader)
        _, channel_count = cls._parse_uint32(bytes_reader)
        _, attachment_count = cls._parse_uint32(bytes_reader)
        _, metadata_count = cls._parse_uint32(bytes_reader)
        _, chunk_count = cls._parse_uint32(bytes_reader)
        _, message_start_time = cls._parse_timestamp(bytes_reader)
        _, message_end_time = cls._parse_timestamp(bytes_reader)
        _, channel_message_counts = cls._parse_map(bytes_reader, 'uint16', 'uint64')

        return StatisticsRecord(
            message_count,
            schema_count,
            channel_count,
            attachment_count,
            metadata_count,
            chunk_count,
            message_start_time,
            message_end_time,
            channel_message_counts
        )


    @classmethod
    def parse_summary_offset(cls, file: BaseReader) -> SummaryOffsetRecord:
        if (record_type := file.read(1)) != b'\x0E':
            raise MalformedMCAP(f'Unexpected record type ({record_type}).')

        _, record_length = cls._parse_uint64(file)
        bytes_reader = BytesReader(file.read(record_length))

        _, group_opcode = cls._parse_uint8(bytes_reader)
        _, group_start = cls._parse_uint64(bytes_reader)
        _, group_length = cls._parse_uint64(bytes_reader)

        return SummaryOffsetRecord(group_opcode, group_start, group_length)
